package edu.kit.ipd.descartes.adaptation.model.adaptation.dmm;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Logger;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.change.ChangeDescription;

import edu.kit.ipd.descartes.adaptation.model.adaptation.IActionHandler;
import edu.kit.ipd.descartes.adaptation.model.adaptation.dmm.util.OclEvaluationHelper;
import edu.kit.ipd.descartes.adaptation.model.adaptation.exceptions.InvalidAdaptationPlan;
import edu.kit.ipd.descartes.adaptation.model.adaptation.exceptions.OperationNotPerformedException;
import edu.kit.ipd.descartes.adaptation.model.repository.dmm.DmlModelLoader;
import edu.kit.ipd.descartes.mm.adaptation.AbstractControlFlowElement;
import edu.kit.ipd.descartes.mm.adaptation.Action;
import edu.kit.ipd.descartes.mm.adaptation.ActionReference;
import edu.kit.ipd.descartes.mm.adaptation.AdaptationPlan;
import edu.kit.ipd.descartes.mm.adaptation.BranchAction;
import edu.kit.ipd.descartes.mm.adaptation.LoopAction;
import edu.kit.ipd.descartes.mm.adaptation.Parameter;
import edu.kit.ipd.descartes.mm.adaptation.StartAction;
import edu.kit.ipd.descartes.mm.adaptation.StopAction;
import edu.kit.ipd.descartes.mm.adaptation.WeightedTactic;
import edu.kit.ipd.descartes.mm.adaptationpoints.AdaptationPoint;
import edu.kit.ipd.descartes.mm.containerrepository.ContainerrepositoryPackage;
import edu.kit.ipd.descartes.mm.resourcelandscape.ResourcelandscapePackage;

public class TacticExecutor {

    private static Logger logger = Logger.getLogger(TacticExecutor.class);
    private List<IObserver> observers = new ArrayList<IObserver>();

    private IActionHandler dmmModelActionHandler;
    private TacticsHistory hist;
    private ChangeDescription changeDescription = null;

    public TacticExecutor(IActionHandler actionExecutor) {
        dmmModelActionHandler = actionExecutor;
        hist = TacticsHistory.getInstance();
        register(BlueYonderExperimentTracker.getInstance()); // TODO implement clean observer
                                                             // pattern
    }

    /**
     * Applies the the given tactic.
     * 
     * @param currentTactic
     *            The tactic to execute
     * @return
     */
    public void applyTactic(WeightedTactic weightedTactic) {
        try {
            logger.info("Applying tactic " + weightedTactic.getUsedTactic().getName() + ", ID: "
                    + weightedTactic.getUsedTactic().getId() + ".");
            DmlModelLoader.startRecording();
            executeAdaptationPlan(weightedTactic.getUsedTactic().getImplementedPlan());
            logger.info("Tactic " + weightedTactic.getUsedTactic().getName() + " successfully applied.");
            hist.add(weightedTactic, TacticsHistory.NO_RESULT, true);

            // if this point is reached everything was fine, so persist them (save the models)
            dmmModelActionHandler.persistActions();
            changeDescription = DmlModelLoader.endRecording();
            return;
        } catch (OperationNotPerformedException e) {
            logger.error("Tactic " + weightedTactic.getUsedTactic().getName()
                    + " could not be executed. Please check adaptation plan and/or log files for errors.", e);
            hist.add(weightedTactic, TacticsHistory.NO_RESULT, false);
            /*
             * TODO Something went wrong. 1) Detect error 2) reevaluate tactics 3) apply different
             * tactic.
             */
        }
    }

    private StartAction findStartAction(AdaptationPlan plan) {
        for (AbstractControlFlowElement abstractControlFlowElement : plan.getSteps()) {
            // Find start of adaptation plan
            if (abstractControlFlowElement instanceof StartAction) {
                return (StartAction) abstractControlFlowElement;
            }
        }
        return null;
    }

    /**
     * Executes the given {@link AdaptationPlan}.
     * 
     * @param plan
     *            The adaptation plan to execute
     * @throws IOException
     * @throws OperationNotPerformedException
     */
    private void executeAdaptationPlan(AdaptationPlan plan) throws OperationNotPerformedException {

        StartAction start = findStartAction(plan);
        if (start == null)
            throw new OperationNotPerformedException("No start action for adaptation plan " + plan.getName()
                    + " found.");

        try {
            executeNextStep(start.getSuccessor());
        } catch (InvalidAdaptationPlan e) {
            logger.error(
                    "Invalid adaptation plan <<" + plan.getName() + ">> when executing action with ID <<"
                            + start.getId() + ">>.", e);
        }
    }

    private void executeNextStep(AbstractControlFlowElement step) throws OperationNotPerformedException,
            InvalidAdaptationPlan {

        if (step == null)
            throw new InvalidAdaptationPlan("Current AbstractControlFlowElement is null!");

        if (step instanceof StopAction) {
            return;
        } else if (step instanceof BranchAction) {
            executeBranchAction(step);
        } else if (step instanceof LoopAction) {
            execLoopAction(step);
        } else if (step instanceof ActionReference) {
            executeReferredAction(step);
        } else
            // Handle invalid actions and control flow elements
            throw new OperationNotPerformedException("No valid control flow element " + step.toString());

        // in any case (except stop action)
        // continue with the next step
        executeNextStep(step.getSuccessor());
    }

    private void executeReferredAction(AbstractControlFlowElement step) throws OperationNotPerformedException {
        ActionReference ref = (ActionReference) step;
        Action action = ref.getRefersTo();
        logger.info("Executing action <<" + action.getName() + ">>.");
        AdaptationPoint currentAdaptationPoint = action.getReferredAdaptationPoint();

        dmmModelActionHandler.execute(currentAdaptationPoint, action.getAdaptationActionOperation());
    }

    private void executeBranchAction(AbstractControlFlowElement branchAction) throws OperationNotPerformedException {
        BranchAction branch = (BranchAction) branchAction;
        if (isBranchConditionTrue(branch.getContext(), branch.getCondition()))
            executeAdaptationPlan(branch.getConditionTrueBranch());
        else
            executeAdaptationPlan(branch.getConditionFalseBranch());
    }

    private void execLoopAction(AbstractControlFlowElement loopAction) throws OperationNotPerformedException {
        LoopAction loop = (LoopAction) loopAction;
        Parameter counter = loop.getCounter();
        int counterValue = Integer.parseInt(counter.getValue());

        int i = 1;

        while (i <= counterValue) {
            logger.debug("Executing iteration " + i + " of loop action " + loop.getBody().getName());
            executeAdaptationPlan(loop.getBody());
            i++;
        }
    }

    private boolean isBranchConditionTrue(EObject obj, String branchCondition) {

        EClass context = null;

        switch (obj.eClass().getClassifierID()) {
        case ResourcelandscapePackage.RUNTIME_ENVIRONMENT:
            context = ResourcelandscapePackage.Literals.RUNTIME_ENVIRONMENT;
            break;
        case ContainerrepositoryPackage.CONTAINER_TEMPLATE:
            context = ContainerrepositoryPackage.Literals.CONTAINER_TEMPLATE;
            break;
        // TODO add further contexts if necessary
        default:
            logger.error("Could not set context for eClass " + obj.getClass());
            return false;
        }

        boolean invariant = OclEvaluationHelper.evaluateOclConstraint(obj, branchCondition, context);
        return invariant;
    }

    public void undoPreviousTactic() {
        logger.info("Reverting model to previous state");
        if (changeDescription != null)
            changeDescription.apply();
    }

    public void register(IObserver observer) {
        observers.add(observer);
    }

    public void remove(IObserver observer) {
        observers.remove(observer);
    }

}
